/* kbtables.c -- keyboard tables and high level access functions
 * Copyright (C) 1995, 1996 Markus F.X.J. Oberhumer
 * For conditions of distribution and use, see copyright notice in kb.h 
 */


#include <stdlib.h>
#include <conio.h>
#include <kb.h>
#include "_kb.h"


/***********************************************************************
//
************************************************************************/

static void init_prefix_tables(void)
{
	int i;
	unsigned char *t;

	t = _kb_prefix_scancode;
	for (i = 0; i < 128; i++)
		t[i] = i;

	/* prefixed keys which map to keypad */
	t[KB_SCAN_SLASH]     = KB_SCAN_SLASH_PAD;
	t[KB_SCAN_ENTER]     = KB_SCAN_ENTER_PAD;

	/* prefixed keys on keypad which map to other keys */
	t[KB_SCAN_ASTERISK_PAD] = KB_SCAN_PRINT;
	t[KB_SCAN_7_PAD]     = KB_SCAN_HOME;
	t[KB_SCAN_8_PAD]     = KB_SCAN_UP;
	t[KB_SCAN_9_PAD]     = KB_SCAN_PGUP;
	t[KB_SCAN_4_PAD]     = KB_SCAN_LEFT;
	t[KB_SCAN_6_PAD]     = KB_SCAN_RIGHT;
	t[KB_SCAN_1_PAD]     = KB_SCAN_END;
	t[KB_SCAN_2_PAD]     = KB_SCAN_DOWN;
	t[KB_SCAN_3_PAD]     = KB_SCAN_PGDN;
	t[KB_SCAN_0_PAD]     = KB_SCAN_INSERT;
	t[KB_SCAN_COMMA_PAD] = KB_SCAN_DEL;

	/* prefixed shift keys */
	t[KB_SCAN_LCONTROL]  = KB_SCAN_RCONTROL;
	t[KB_SCAN_ALT]       = KB_SCAN_ALTGR;
	t[KB_SCAN_LSHIFT]    = KB_SCAN_LSHIFT_VIRTUAL;
	t[KB_SCAN_RSHIFT]    = KB_SCAN_RSHIFT_VIRTUAL;

	/* prefixed special keys */
	t[KB_SCAN_NUMLOCK]   = KB_SCAN_PAUSE_VIRTUAL;
	t[KB_SCAN_SCRLOCK]   = KB_SCAN_CONTROL_BREAK_VIRTUAL;


	t = _kb_inverse_prefix_scancode;
	for (i = 0; i < 128; i++)
		t[i] = i;
	for (i = 0; i < 128; i++)
		if (_kb_prefix_scancode[i] != i)
			t[_kb_prefix_scancode[i]] = i;
}


static void init_shift_state_tables(void)
{
	int i;
	unsigned short *t;

	t = _kb_shift_state_table;
	for (i = 0; i < 128; i++)
		t[i] = 0;

/* shift keys */
	t[KB_SCAN_LSHIFT]          = KB_SHIFT_LSHIFT | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_RSHIFT]          = KB_SHIFT_RSHIFT | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_LCONTROL]        = KB_SHIFT_LCONTROL | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_RCONTROL]        = KB_SHIFT_RCONTROL | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_ALT]             = KB_SHIFT_ALT | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_ALTGR]           = KB_SHIFT_ALTGR | KB_SHIFT_VIRTUAL;
/* these are toggles and therefore not in the unshift table */
	t[KB_SCAN_CAPSLOCK]        = KB_SHIFT_CAPSLOCK | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_NUMLOCK]         = KB_SHIFT_NUMLOCK | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_SCRLOCK]         = KB_SHIFT_SCRLOCK | KB_SHIFT_VIRTUAL;
/* these are toggles that DO produce a keypress */
	t[KB_SCAN_INSERT]          = KB_SHIFT_INSERT;
/* these are toggles for virtual keys */
	t[KB_SCAN_PAUSE_VIRTUAL]   = KB_SHIFT_PAUSE | KB_SHIFT_VIRTUAL;
	t[KB_SCAN_CONTROL_BREAK_VIRTUAL] =
		KB_SHIFT_CONTROL_BREAK | KB_SHIFT_VIRTUAL;
/* these are used to indicate that the key is a shift key */
/* and therfore no keypress should be stored */
	t[KB_SCAN_LSHIFT_VIRTUAL]  = KB_SHIFT_VIRTUAL;
	t[KB_SCAN_RSHIFT_VIRTUAL]  = KB_SHIFT_VIRTUAL;
	t[KB_SCAN_LCONTROL_VIRTUAL] = KB_SHIFT_VIRTUAL;


	t = _kb_unshift_state_table;
	for (i = 0; i < 128; i++)
		t[i] = 0;
	t[KB_SCAN_LSHIFT]          = KB_SHIFT_LSHIFT;
	t[KB_SCAN_RSHIFT]          = KB_SHIFT_RSHIFT;
	t[KB_SCAN_LCONTROL]        = KB_SHIFT_LCONTROL;
	t[KB_SCAN_RCONTROL]        = KB_SHIFT_RCONTROL;
	t[KB_SCAN_ALT]             = KB_SHIFT_ALT;
	t[KB_SCAN_ALTGR]           = KB_SHIFT_ALTGR;

/* negate for use with & (and) */
	for (i = 0; i < 128; i++)
		t[i] = ~t[i];
}


/***********************************************************************
// These tables are hardcoded for American keyboards.
// Note the additional entries for the mapped scan codes 
// in the last rows.
************************************************************************/

/* unshifted ASCII for scan codes */
char _kb_asc_table[128] =
{
/*	0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F           */
	0  ,27 ,'1','2','3','4','5','6','7','8','9','0','-','=',8  ,9  ,   /* 0 */
	'q','w','e','r','t','y','u','i','o','p','[',']',13 ,0  ,'a','s',   /* 1 */
	'd','f','g','h','j','k','l',';',39 ,'`',0  ,92 ,'z','x','c','v',   /* 2 */
	'b','n','m',',','.','/',0  ,'*',0  ,' ',0  ,0  ,0  ,0  ,0  ,0  ,   /* 3 */
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,'7','8','9','-','4','5','6','+','1',   /* 4 */
	'2','3','0',',',0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,   /* 5 */
	13 ,0  ,'/',0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,127,   /* 6 */
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0      /* 7 */
};

/* shifted ASCII for scan codes */
char _kb_shift_table[128] =
{
/*	0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F           */
	0  ,27 ,'!','@','#','$','%','^','&','*','(',')','_','+',8  ,9  ,   /* 0 */
	'Q','W','E','R','T','Y','U','I','O','P','{','}',13 ,0  ,'A','S',   /* 1 */
	'D','F','G','H','J','K','L',':',34 ,'~',0  ,'|','Z','X','C','V',   /* 2 */
	'B','N','M','<','>','?',0  ,'*',0  ,' ',0  ,0  ,0  ,0  ,0  ,0  ,   /* 3 */
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,'7','8','9','-','4','5','6','+','1',   /* 4 */
	'2','3','0',',',0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,   /* 5 */
	13 ,0  ,'/',0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,127,   /* 6 */
	0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0  ,0      /* 7 */
};

static void init_ascii_tables(void)
{
	if (!(_kb_flags & KB_FLAG_LOCALIZE))
		return;

	/* FIXME: use local scancode-to-ASCII conversions */
}


static unsigned ascii_code(unsigned k)
{
	unsigned key;
	unsigned scan = k & 0x7f;

	if (k & ((KB_SHIFT_ANY_SHIFT | KB_SHIFT_CAPSLOCK) << 8))
		key = _kb_shift_table[scan];
	else
		key = _kb_asc_table[scan];

	return key;
}


/***********************************************************************
// FIXME: these tables need to be reworked
************************************************************************/

unsigned char _kb_alt_table[128];
unsigned char _kb_control_table[128];

static void init_getch_tables(void)
{
	int i;
	unsigned char *t;

	t = _kb_alt_table;
	for (i = 0; i < 128; i++)
		t[i] = _kb_inverse_prefix_scancode[i];
	for (i = KB_SCAN_HOME; i <= KB_SCAN_DEL; i++)
		t[i] += 0x50;
	t[KB_SCAN_ENTER_PAD] = 0xa6;

	t = _kb_control_table;
	for (i = 0; i < 128; i++)
		t[i] = _kb_inverse_prefix_scancode[i];
}


static unsigned getch_code(unsigned k)
{
	unsigned key;
	unsigned scan = k & 0x7f;

	if (k & (KB_SHIFT_ANY_ALT << 8))
		key = _kb_alt_table[scan];
	else if (k & (KB_SHIFT_ANY_CONTROL << 8))
		key = _kb_control_table[scan];
	else
		key = _kb_inverse_prefix_scancode[scan];

	return key;
}


/***********************************************************************
//
************************************************************************/

void _kb_init_tables(void)
{
	static unsigned char done = 0;
	if (done)
		return;
	done = 1;

	init_prefix_tables();			/* must be initialized first */
	init_shift_state_tables();
	init_ascii_tables();
	init_getch_tables();
}


/***********************************************************************
// this is the preferred access function
// it returns the kb_keypress() information in the upper 16 bits.
************************************************************************/

unsigned long kb_inkey(void)
{
	unsigned key;
	unsigned long k;

	if (!_kb_mode)
	{
		k = getch();
		if (!k)
			k = ((unsigned long)(getch())) << 8;
		return k;
	}

	key = kb_keypress();
	if (!key)
		return 0;			/* no key pressed */

	k  = ascii_code(key);
	k |= getch_code(key) << 8;
	k |= (unsigned long) key << 16;

	return k;
}


/***********************************************************************
// mostly compatible to getch()
// FIXME: I want to make this 100% compatible to Borland C getch()
************************************************************************/

int kb_getch(void)
{
	static int last_key = 0;
	int key, k;

	if (!_kb_mode)
		return getch();

	if (last_key)
	{
		key = last_key;
		last_key = 0;
		return key;
	}

	key = kb_keypress();
	if (!key)
		return -1;			/* no key pressed */

	k = ascii_code(key);
	if (k)
		return k;
	else
	{
		last_key = getch_code(key);
		return 0;
	}
}


/*
vi:ts=4
*/
